#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <locale.h>
#include <utmp.h>

#include "unsecure.h"

#define LOGIN_PATH "/usr/local/bin:/usr/bin:/bin"
#define DEFAULT_SHELL "/bin/sh"
#define MAX_NAME_LENGTH 20
#define KEY_SIZE 8

#define INTRO_TEXT "Unsecure login (c) 1999 Nicolas GEORGE\n\n"
#define FAILURE_TIME 3000000

extern char **environ;

char *pname;

const char *basename(const char *n)
{
  const char *b;

  b=n;
  while(*n){
    if(*n== '/')
        b=n+1;
      n++;
  }
  return(b);
}

static char *addstr(char *t,const char *s)
{
  for(;*s;s++,t++)
    *t=*s;
  *t=0;
  return(t);
}

static char *add_env(char *b,char *k,char *v)
{
  char *c;

  c=addstr(b,k);
  *(c++)='=';
  c=addstr(c,v);
  *(c++)=0;
  putenv(b);
  return(c);
}

static void modify_environment(const struct passwd *pw)
{
  char *term;
  char *b;
  int l=10+4+5+4+7+4;

  term=getenv("TERM");
  if(term!=NULL){
    l+=strlen(term+6);
  }
  environ=malloc(2*sizeof(char *));
  environ[0]=0;
  l+=strlen(pw->pw_dir);
  l+=strlen(pw->pw_shell);
  l+=strlen(pw->pw_name);
  l+=strlen(pw->pw_name);
  l+=strlen(LOGIN_PATH);
  b=malloc(l);
  if(term!=NULL)
    b=add_env(b,"TERM",term);
  b=add_env(b,"HOME",pw->pw_dir);
  b=add_env(b,"SHELL",pw->pw_shell);
  b=add_env(b,"USER",pw->pw_name);
  b=add_env(b,"LOGNAME",pw->pw_name);
  b=add_env(b,"PATH",LOGIN_PATH);
}

static void run_shell(const char *shell)
{
  char **args;
  const char *sbn;

  if((shell==NULL)|(*shell==0))
    shell=DEFAULT_SHELL;

  args=malloc(sizeof(char *)*2);
  
  sbn=basename(shell);
  args[0]=malloc(strlen(sbn)+2);
  args[0][0]='-';
  strcpy(args[0]+1,sbn);
  args[1]=NULL;
  execv(shell,(char **)args);
  perror("Unsecure: cannot run shell");
}

void log_message(int s,char *l)
{
  openlog(basename(pname),0,LOG_AUTH);
  syslog(LOG_NOTICE,"unsecure login %s to %s",s?"success":"failure",l);
  closelog();
}

static void log_utmp(const struct passwd *pw)
{
  struct utmp ut,*rut;
  pid_t pid;

  setutent();

  pid=getpid();
  while(1){
    rut=getutent();
    if(rut==NULL){
      fprintf(stderr,"Unsecure: not in utmp\n");
      return;
    }
    if((rut->ut_type==USER_PROCESS)&&(rut->ut_pid==pid))
      break;
  }

  ut=*rut;
  strcpy(ut.ut_user,pw->pw_name);
  pututline(&ut);
  printf("%s\n",pw->pw_name);
  setutent();

  endutent();
}

int main(int argc,char **argv)
{
  char un[MAX_NAME_LENGTH+1];
  char kin[KEY_SIZE+1],kou[128],*tkou;
  int i=0;
  int c;
  struct passwd *pw;
  FILE *ran;
  char *ttyn;

  pname=argv[0];
  setlocale(LC_ALL,"");

  printf(INTRO_TEXT);
  printf("Secure login: ");
  fflush(stdout);

  while(1){
    c=getchar();
    if(c==EOF)
      return(1);
    if(c=='\n')
      break;
    if(i<MAX_NAME_LENGTH)
      un[i++]=c;
  }
  un[i]=0;

  ran=fopen("/dev/random","rb");
  if(ran==NULL){
    perror("Unsecure: random device");
    return(1);
  }
  for(i=0;i<KEY_SIZE;i++){
    c=63;
    while((c==62)||(c==63))
      c=getc(ran)&63;
    kin[i]=c<10?c+'0':(c<36?c+('A'-10):c+('a'-36));
  }
  fclose(ran);
  kin[i]=0;

  printf("%s-",kin);
  fflush(stdout);
  tkou=getpass("Password: ");
  for(i=0;tkou[i]!=0;i++)
    kou[i]=tkou[i];
  kou[i]=0;

  for(i=0;i<nutils;i++){
    if(strcmp(un,utils[i].n)==0)
      break;
  }

  if(i==nutils){
    printf("Invalid password\n");
    log_message(0,un);
    signal(SIGINT,SIG_IGN);
    signal(SIGTSTP,SIG_IGN);
    usleep(FAILURE_TIME);
    return(1);
  }

  pw=getpwuid(utils[i].u);
  endpwent();
  if(pw==0){
    perror("Unsecure: internal error");
    return(1);
  }
  log_utmp(pw);

#if(0)
  if(initgroups(pw->pw_name,pw->pw_gid)==-1){
    perror("Unsecure: cannot set groups");
    return(1);
  }
  endgrent();
#endif

  ttyn=ttyname(STDIN_FILENO);
  if(ttyn==NULL){
    perror("Unsecure: cannot get tty name");
    return(1);
  }
  if(chown(ttyn,pw->pw_uid,pw->pw_gid))
    perror("Unsecure: cannot change tty ownership");

  if(setgid(pw->pw_gid)){
    perror("Unsecure: cannot set group id");
    return(1);
  }
  if(setegid(pw->pw_gid)){
    perror("Unsecure: cannot set effective group id");
    return(1);
  }
  if(setuid(pw->pw_uid)){
    perror("Unsecure: cannot set user id");
    return(1);
  }
  if(seteuid(pw->pw_uid)){
    perror("Unsecure: cannot set effective user id");
    return(1);
  }

  if(!utils[i].f(kin,kou)){
    printf("Invalid password\n");
    log_message(0,un);
    signal(SIGINT,SIG_IGN);
    signal(SIGTSTP,SIG_IGN);
    usleep(FAILURE_TIME);
    return(1);
  }
  log_message(1,un);

  modify_environment(pw);
  if (chdir(pw->pw_dir)){
    perror("Unsecure: home directory");
    return(1);
  }

  run_shell(pw->pw_shell);
  return(1);
}
